/* David Leonard, 2003. Public domain. */
/* $Id: Stack.c,v 1.5 2003/04/27 14:40:57 d Exp $ */

/*
 * A widget to display a stack of labels.
 * Each label should have a field
 */

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xmu/Converters.h>
#include <X11/Shell.h>

#define ALLOCATE_LOCAL_FALLBACK(_size) XtMalloc(_size)
#define DEALLOCATE_LOCAL_FALLBACK(_ptr) XtFree(_ptr)
#include <X11/Xalloca.h>

#include "CardP.h"
#include "StackP.h"

#ifdef DEBUG
# define HERE()	fprintf(stderr, "%s()\n", __FUNCTION__)
#else
# define HERE()	/* nothing */
#endif

/* 
 * Prototypes
 */
static void	StackGetGC(StackWidget);
static void	StackGetPixmapSize(StackWidget, int);
static void	StackDrawPixmap(StackWidget, CardID, int, int, int, int,
				int, int, Region);
static void	StackPartOrigin(StackWidget, Cardinal, Position *, Position *);
static void	StackExtent(StackWidget, Dimension, Dimension, 
				Dimension *, Dimension *);
static Boolean	PointInRect(RectObj, Position, Position);
static Boolean	StackCallSplitCallback(StackWidget, Cardinal);
static Boolean	StackCallDropCallback(StackWidget, CardID *, Cardinal, 
				StackWidget);
static Cardinal StackCount(StackWidget);
static CardObject StackGetCard(StackWidget, int);
static void	StackGetBottomCentre(StackWidget, Position *, Position *);
static Cardinal StackCountSelected(Widget, Position, Position);

static void	StartDrag(Widget, XEvent *, String *, Cardinal *);
static void	AbortDrag(Widget, XEvent *, String *, Cardinal *);
static void	Hit(Widget, XEvent *, String *, Cardinal *);
static Boolean	PopupEventHandler(XEvent *);
static void	PopupStopDrag(XEvent *, Boolean);
static Widget	FindWidget(Widget, int, int);
static Boolean	PopupCanEnter(Widget, XEvent *);
static void	PopupEnter(Widget, XEvent *);
static void	PopupLeave(Widget, XEvent *);
static Boolean	PopupCanDropCards(Widget);

/*
 * Macros
 */
#define CARDID_TO_INDEX(card)	(CARD_IS_DOWN(card) \
					? _CARD_BACK \
					: (card) & CARD_MASK)
#define StackSuperclass(w) ((CompositeWidgetClass)XtSuperclass(w))
#define XtX(w)            (((RectObj)w)->rectangle.x)
#define XtY(w)            (((RectObj)w)->rectangle.y)
#define XtWidth(w)        (((RectObj)w)->rectangle.width)
#define XtHeight(w)       (((RectObj)w)->rectangle.height)
#define XtBorderWidth(w)  (((RectObj)w)->rectangle.border_width)

/*
 * Class Methods
 */
static void StackClassInitialize(void);
static void StackInitialize(Widget, Widget, ArgList, Cardinal *);
static void StackDestroy(Widget);
static void StackChangeManaged(Widget);
static void StackInsertChild(Widget);
static void StackExpose(Widget, XEvent *, Region);
static void StackResize(Widget);
static void StackRealize(Widget, XtValueMask *, XSetWindowAttributes *);
static Boolean StackSetValues(Widget, Widget, Widget, ArgList, Cardinal *);
static XtGeometryResult StackQueryGeometry(Widget, XtWidgetGeometry *, 
							XtWidgetGeometry *);

/* 
 * Actions
 */
static void Highlight(Widget, XEvent *, String *, Cardinal *);
static void Unhighlight(Widget, XEvent *, String *, Cardinal *);

/*
 * Resources
 */
#define offset(field)	XtOffsetOf(StackRec, stack.field)
static XtResource resources[] = {
/*	{ name, class, type,
	  size, offset,
	  default_type, default_addr }, */
	{ XtNforeground, XtCForeground, XtRPixel,
	  sizeof(Pixel), offset(foreground),
	  XtRString, XtDefaultForeground },
	{ XtNhighlight, XtCForeground, XtRPixel,
	  sizeof(Pixel), offset(highlight),
	  XtRString, XtDefaultForeground },
	{ XtNdropCallback, XtCCallback, XtRPointer,
	  sizeof(XtPointer), offset(drop_callback),
	  XtRImmediate, (XtPointer)StackAlwaysDrop },
	{ XtNsplitCallback, XtCCallback, XtRPointer,
	  sizeof(XtPointer), offset(split_callback),
	  XtRImmediate, (XtPointer)StackAlwaysSplit },
	{ XtNhitNotify, XtCCallback, XtRCallback,
	  sizeof(XtPointer), offset(hit_notify),
	  XtRImmediate, (XtPointer)NULL },
	{ XtNdropNotify, XtCCallback, XtRCallback,
	  sizeof(XtPointer), offset(drop_notify),
	  XtRImmediate, (XtPointer)NULL },
	{ XtNgravity, XtCGravity, XtRGravity,
	  sizeof(XtGravity), offset(gravity),
	  XtRImmediate, (XtPointer)CenterGravity },
	{ XtNvStepMin, XtCVStepMin, XtRDimension,
	  sizeof(Dimension), offset(v_step_min),
	  XtRImmediate, (XtPointer)8 },
	{ XtNvStepMax, XtCVStepMax, XtRDimension,
	  sizeof(Dimension), offset(v_step_max),
	  XtRImmediate, (XtPointer)24 },
	{ XtNhStepMin, XtCHStepMin, XtRDimension,
	  sizeof(Dimension), offset(h_step_min),
	  XtRImmediate, (XtPointer)8 },
	{ XtNhStepMax, XtCHStepMax, XtRDimension,
	  sizeof(Dimension), offset(h_step_max),
	  XtRImmediate, (XtPointer)16 },
	{ XtNhighlightWidth, XtCWidth, XtRDimension,
	  sizeof(Dimension), offset(highlight_width),
	  XtRImmediate, (XtPointer)1 },
	{ XtNhighlightHeight, XtCHeight, XtRDimension,
	  sizeof(Dimension), offset(highlight_height),
	  XtRImmediate, (XtPointer)1 },
	{ XtNcompact, XtCCompact, XtRBoolean,
	  sizeof(Boolean), offset(compact),
	  XtRImmediate, (XtPointer)False },

#define CARD_RESOURCE(name, card) \
	{ name, XtCCardPixmap, XtRPixmap, \
	  sizeof(Pixmap), offset(card_pixmap[card]), \
	  XtRPixmap, (XtPointer)None }

	CARD_RESOURCE(XtNplaceholderPixmap, _CARD_PLACE),
	CARD_RESOURCE(XtNbackPixmap, 	    _CARD_BACK),
	CARD_RESOURCE(XtNaceDiamondsPixmap, SUIT_DIAMONDS | FACE_ACE),
	CARD_RESOURCE(XtNtwoDiamondsPixmap, SUIT_DIAMONDS | 2),
	CARD_RESOURCE(XtNthreeDiamondsPixmap, SUIT_DIAMONDS | 3),
	CARD_RESOURCE(XtNfourDiamondsPixmap, SUIT_DIAMONDS | 4),
	CARD_RESOURCE(XtNfiveDiamondsPixmap, SUIT_DIAMONDS | 5),
	CARD_RESOURCE(XtNsixDiamondsPixmap, SUIT_DIAMONDS | 6),
	CARD_RESOURCE(XtNsevenDiamondsPixmap, SUIT_DIAMONDS | 7),
	CARD_RESOURCE(XtNeightDiamondsPixmap, SUIT_DIAMONDS | 8),
	CARD_RESOURCE(XtNnineDiamondsPixmap, SUIT_DIAMONDS | 9),
	CARD_RESOURCE(XtNtenDiamondsPixmap, SUIT_DIAMONDS | 10),
	CARD_RESOURCE(XtNjackDiamondsPixmap, SUIT_DIAMONDS | FACE_JACK),
	CARD_RESOURCE(XtNqueenDiamondsPixmap, SUIT_DIAMONDS | FACE_QUEEN),
	CARD_RESOURCE(XtNkingDiamondsPixmap, SUIT_DIAMONDS | FACE_KING),
	CARD_RESOURCE(XtNaceHeartsPixmap, SUIT_HEARTS | FACE_ACE),
	CARD_RESOURCE(XtNtwoHeartsPixmap, SUIT_HEARTS | 2),
	CARD_RESOURCE(XtNthreeHeartsPixmap, SUIT_HEARTS | 3),
	CARD_RESOURCE(XtNfourHeartsPixmap, SUIT_HEARTS | 4),
	CARD_RESOURCE(XtNfiveHeartsPixmap, SUIT_HEARTS | 5),
	CARD_RESOURCE(XtNsixHeartsPixmap, SUIT_HEARTS | 6),
	CARD_RESOURCE(XtNsevenHeartsPixmap, SUIT_HEARTS | 7),
	CARD_RESOURCE(XtNeightHeartsPixmap, SUIT_HEARTS | 8),
	CARD_RESOURCE(XtNnineHeartsPixmap, SUIT_HEARTS | 9),
	CARD_RESOURCE(XtNtenHeartsPixmap, SUIT_HEARTS | 10),
	CARD_RESOURCE(XtNjackHeartsPixmap, SUIT_HEARTS | FACE_JACK),
	CARD_RESOURCE(XtNqueenHeartsPixmap, SUIT_HEARTS | FACE_QUEEN),
	CARD_RESOURCE(XtNkingHeartsPixmap, SUIT_HEARTS | FACE_KING),
	CARD_RESOURCE(XtNaceClubsPixmap, SUIT_CLUBS | FACE_ACE),
	CARD_RESOURCE(XtNtwoClubsPixmap, SUIT_CLUBS | 2),
	CARD_RESOURCE(XtNthreeClubsPixmap, SUIT_CLUBS | 3),
	CARD_RESOURCE(XtNfourClubsPixmap, SUIT_CLUBS | 4),
	CARD_RESOURCE(XtNfiveClubsPixmap, SUIT_CLUBS | 5),
	CARD_RESOURCE(XtNsixClubsPixmap, SUIT_CLUBS | 6),
	CARD_RESOURCE(XtNsevenClubsPixmap, SUIT_CLUBS | 7),
	CARD_RESOURCE(XtNeightClubsPixmap, SUIT_CLUBS | 8),
	CARD_RESOURCE(XtNnineClubsPixmap, SUIT_CLUBS | 9),
	CARD_RESOURCE(XtNtenClubsPixmap, SUIT_CLUBS | 10),
	CARD_RESOURCE(XtNjackClubsPixmap, SUIT_CLUBS | FACE_JACK),
	CARD_RESOURCE(XtNqueenClubsPixmap, SUIT_CLUBS | FACE_QUEEN),
	CARD_RESOURCE(XtNkingClubsPixmap, SUIT_CLUBS | FACE_KING),
	CARD_RESOURCE(XtNaceSpadesPixmap, SUIT_SPADES | FACE_ACE),
	CARD_RESOURCE(XtNtwoSpadesPixmap, SUIT_SPADES | 2),
	CARD_RESOURCE(XtNthreeSpadesPixmap, SUIT_SPADES | 3),
	CARD_RESOURCE(XtNfourSpadesPixmap, SUIT_SPADES | 4),
	CARD_RESOURCE(XtNfiveSpadesPixmap, SUIT_SPADES | 5),
	CARD_RESOURCE(XtNsixSpadesPixmap, SUIT_SPADES | 6),
	CARD_RESOURCE(XtNsevenSpadesPixmap, SUIT_SPADES | 7),
	CARD_RESOURCE(XtNeightSpadesPixmap, SUIT_SPADES | 8),
	CARD_RESOURCE(XtNnineSpadesPixmap, SUIT_SPADES | 9),
	CARD_RESOURCE(XtNtenSpadesPixmap, SUIT_SPADES | 10),
	CARD_RESOURCE(XtNjackSpadesPixmap, SUIT_SPADES | FACE_JACK),
	CARD_RESOURCE(XtNqueenSpadesPixmap, SUIT_SPADES | FACE_QUEEN),
	CARD_RESOURCE(XtNkingSpadesPixmap, SUIT_SPADES | FACE_KING),
	CARD_RESOURCE(XtNjokerRedPixmap, SUIT_SPECIAL | COLOUR_RED | FACE_JOKER),
	CARD_RESOURCE(XtNjokerBlackPixmap, SUIT_SPECIAL | COLOUR_BLACK | FACE_JOKER),
};
#undef offset
#undef CARD_RESOURCE

static XtActionsRec actions[] = {
/*	{ name,			procedure }, */
	{ "highlight",		Highlight },
	{ "unhighlight",	Unhighlight },
	{ "start-drag",		StartDrag },
	{ "abort-drag",		AbortDrag },
	{ "hit",		Hit },
};

static char translations[] =
"<Enter>:"		"highlight()\n"
"<Leave>:"		"unhighlight()\n"
"<Btn1Down>:"		"start-drag()\n"
"<Btn1Up>:"		"hit()\n"
;

/* 
 * Allow non-widgets to be added to stacks
 */
static CompositeClassExtensionRec compositeClassExtension = {
	NULL,					/* next_extension */
	NULLQUARK,				/* record_type */
	XtCompositeExtensionVersion,		/* version */
	sizeof (CompositeClassExtensionRec),	/* record_size */
	TRUE,					/* accepts_objects */
	FALSE,					/* allows_change_managed_set */
};

/*
 * Class record (methods and properties)
 */
StackClassRec stackClassRec = {
    /* core */
    {
	(WidgetClass)&compositeClassRec,	/* superclass */
	"Stack",				/* class_name */
	sizeof (StackRec),			/* widget_size */
	StackClassInitialize,			/* class_initialize */
	NULL,					/* class_part_initialize */
	False,					/* class_inited */
	StackInitialize,			/* initialize */
	NULL,					/* initialize_hook */
	StackRealize,				/* realize */
	actions,				/* actions */
	XtNumber(actions),			/* num_actions */
	resources,				/* resources */
	XtNumber(resources),			/* num_resources */
	NULLQUARK,				/* xrm_class */
	True,					/* compress_motion */
	True,					/* compress_exposure */
	True,					/* compress_enterleave */
	False,					/* visible_interest */
	StackDestroy,				/* destroy */
	StackResize,				/* resize */
	StackExpose,				/* expose */
	StackSetValues,				/* set_values */
	NULL,					/* set_values_hook */
	XtInheritSetValuesAlmost,		/* set_values_almost */
	NULL,					/* get_values_hook */
	NULL,					/* accept_focus */
	XtVersion,				/* version */
	NULL,					/* callback_private */
	translations,				/* tm_table */
	StackQueryGeometry,			/* query_geometry */
	XtInheritDisplayAccelerator,		/* display_accelerator */
	NULL,					/* extension */
    },
    /* composite */
    {
	NULL,					/* geometry_manager */
	StackChangeManaged,			/* change_managed */
	StackInsertChild,			/* insert_child */
	XtInheritDeleteChild,			/* delete_child */
	(XtPointer)&compositeClassExtension,	/* extension */
    },
    /* stack */
    {
	NULL,					/* extension */
    }
};

WidgetClass stackWidgetClass = (WidgetClass)&stackClassRec;

/*------------------------------------------------------------
 * Convenience API for card games.
 */

/*
 * Return the number of cards on the stack
 */
Cardinal
StackDepth(Widget w)
{
	StackWidget sw = (StackWidget)w;

	HERE();
	XtCheckSubclass(w, stackWidgetClass, "StackDepth: bad argument");
	return StackCount(sw);
}

/*
 * Return the card at the given depth.
 * The top card has a depth of 0.
 * Returns CARD_NONE if there is no card at that depth.
 */
CardID
StackPeek(Widget w, Cardinal depth)
{
	StackWidget sw = (StackWidget)w;
	int i;

	HERE();
	XtCheckSubclass(w, stackWidgetClass, "StackPeek: bad argument");
	i = StackCount(sw) - 1 - depth;
	if (i < 0)
		return CARD_NONE;
	return StackGetCard(sw, i)->card.card_id;
}

/*
 * Push a card on top of the stack.
 * Creates a new Card object.
 * Pushing CARD_NONE results in a warning message.
 */
void
StackPush(Widget w, CardID card)
{
	Arg args[5];
	int argc;

	HERE();
	XtCheckSubclass(w, stackWidgetClass, "StackPush: bad argument");
	if (CARDID_TO_INDEX(card) == CARD_NONE) {
#ifdef DEBUG
		XtWarning("StackPush: pushing CARD_NONE ignored");
#endif
	} else {
		argc = 0;
		XtSetArg(args[argc], XtNcardID, card); argc++;
		XtCreateManagedWidget("card", cardObjectClass, w, args, argc);
	}
}

/* 
 * Return the card on top of the stack, and destroy
 * its representative widget.
 * Returns CARD_NONE if there are no more cards to pop.
 * Destroys the card object popped.
 */
CardID
StackPop(Widget w)
{
	StackWidget sw = (StackWidget)w;
	CardObject c;
	CardID card;
	int depth;

	HERE();
	XtCheckSubclass(w, stackWidgetClass, "StackPop: bad argument");
	depth = StackCount(sw);
	if (depth == 0)
		card = CARD_NONE;
	else {
		c = StackGetCard(sw, depth - 1);
		card = c->card.card_id;
		XtUnmanageChild((Widget)c);
		XtDestroyWidget((Widget)c);
	}
	return card;
}

/*
 * Remove everything from the stack
 */
void
StackClear(Widget w)
{
	StackWidget sw = (StackWidget)w;
	int count = StackCount(sw);

	while (count--)
		StackPop(w);
}

/*
 * Given a widget-relative coordinate of a mouse click,
 * return the number of top cards on top of the stack
 * that have been selected. Returns 0 if the stack is
 * empty.
 *
 * (A position on a card selects that card and the cards above it)
 */
static Cardinal
StackCountSelected(Widget w, Position x, Position y)
{
	StackWidget sw = (StackWidget)w;
	CardObject co;
	int i;

	HERE();
	XtCheckSubclass(w, stackWidgetClass, "StackCountSelected: bad argument");

	for (i = 1; i <= sw->composite.num_children; i++) {
		co = (CardObject)sw->composite.
			children[sw->composite.num_children - i];
		if (XtIsManaged((Widget)co) && PointInRect((RectObj)co, x, y))
			return i;
	}
	return 0;
}

/*
 * Move the top 'count' cards from the 'src' stack to the 
 * 'dst' stack, keeping their order.
 */
void
StackMove(Widget dst, Widget src, Cardinal count)
{
	CardID *cards;
	int i;

	HERE();

	XtCheckSubclass(dst, stackWidgetClass, "StackMove: bad argument");
	if (count > StackDepth(src))
		XtError("StackMove: count exceeds available cards");

	/* Allocate local storage for the CardIDs */
	cards = (CardID *)ALLOCATE_LOCAL(count * sizeof (CardID));

	for (i = 0; i < count; i++) 
		cards[i] = StackPop(src);
	for (i = count - 1; i >= 0; i--) 
		StackPush(dst, cards[i]);

	DEALLOCATE_LOCAL(cards);
}

extern Boolean 
StackNeverSplitProc(source, closure, depth, check_only)
        Widget source;
        XtPointer closure;
        Cardinal depth;
        Boolean check_only;
{
	return False;
}

extern Boolean 
StackNeverDropProc(target, closure, cards, num_cards, source, check_only)
        Widget target;
        XtPointer closure;
        CardID *cards;
        Cardinal num_cards;
        Widget source;
        Boolean check_only;
{
	return False;
}


/*------------------------------------------------------------
 * Helper functions
 */

/*
 * return the number of managed cards in the stack
 */
static Cardinal
StackCount(StackWidget sw)
{
	int i;
	Cardinal depth = 0;

	for (i = 0; i < sw->composite.num_children; i++)
		if (XtIsManaged(sw->composite.children[i]))
			depth++;
	return depth;
}

/*
 * return the 'index'th managed card in the stack
 */
static CardObject
StackGetCard(StackWidget sw, int index)
{
	int i;
	for (i = 0; i < sw->composite.num_children; i++) {
		if (!XtIsManaged(sw->composite.children[i]))
			continue;
		if (index-- == 0)
			return (CardObject)sw->composite.children[i];
	}
	return NULL;
}

/*
 * Return true if (x,y) is inside the rectangle object
 * [[Could have used the INBOX() macro from <X11/region.h>?]]
 */
static Boolean
PointInRect(RectObj r, Position x, Position y)
{
	return x >= r->rectangle.x && x < r->rectangle.x + r->rectangle.width
	    && y >= r->rectangle.y && y < r->rectangle.y + r->rectangle.height;
}

/*
 * Obtain a graphics context for drawing the cards
 */
static void
StackGetGC(sw)
	StackWidget sw;
{
	XtGCMask mask = 0;
	XGCValues values;

	HERE();

	values.foreground = sw->stack.foreground;	mask |= GCForeground;
	values.background = sw->core.background_pixel;	mask |= GCBackground;
	values.graphics_exposures = False; 	mask |= GCGraphicsExposures;
	if (sw->stack.gc)
		XtReleaseGC((Widget)sw, sw->stack.gc);
	sw->stack.gc = XtGetGC((Widget)sw, mask, &values);

	values.foreground = sw->stack.highlight;
	if (sw->stack.highlight_gc)
		XtReleaseGC((Widget)sw, sw->stack.highlight_gc);
	sw->stack.highlight_gc = XtGetGC((Widget)sw, mask, &values);

}

/*
 * Read and cache the dimensions of a pixmap
 */
static void
StackGetPixmapSize(sw, index)
	StackWidget sw;
	int index;
{
	Pixmap pixmap = sw->stack.card_pixmap[index];
	unsigned int bw;
	int x, y;
	unsigned int w, h, depth;
	Window root;

	HERE();
	if (sw->stack.card_depth[index] == -1) {
		if (pixmap && XGetGeometry(XtDisplay(sw),
			pixmap, &root, &x, &y, &w, &h, &bw, &depth))
		{
			sw->stack.card_width[index] = w;
			sw->stack.card_height[index] = h;
			sw->stack.card_depth[index] = depth;
		} else {
			char msg[80];

			snprintf(msg, sizeof msg, 
			    "Missing pixmap for cardID 0x%x", index);
			XtWarning(msg);
			sw->stack.card_width[index] = 0;
			sw->stack.card_height[index] = 0;
			sw->stack.card_depth[index] = 0;
			abort();
		}
	}
}

/*
 * Draw one card's pixmap at the given position
 */
static void
StackDrawPixmap(sw, index, sx, sy, w, h, x, y, region)
	StackWidget sw;
	CardID index;
	int sx, sy, w, h, x, y;
	Region region;
{
	Pixmap pixmap;

	HERE();
	if (region && XRectInRegion(region, x, y, w, h) == RectangleOut)
		return;
	if (sw->stack.card_depth[index] == 0)	/* bad pixmap */
		return;
	pixmap = sw->stack.card_pixmap[index];
	if (sw->stack.card_depth[index] == 1)
		XCopyPlane(XtDisplay(sw), pixmap, XtWindow(sw),
			sw->stack.gc, sx, sy, w, h, x, y, 1L);
	else
		XCopyArea(XtDisplay(sw), pixmap, XtWindow(sw),
			sw->stack.gc, sx, sy, w, h, x, y);
}

/*
 * Compute the top left coordinate of a card group
 */
static void
StackPartOrigin(sw, depth, x_return, y_return)
	StackWidget sw;
	Cardinal depth;
	Position *x_return, *y_return;
{
	Position x, y;

	switch (sw->stack.gravity) {
	case NorthGravity:
		x = 0;
		y = sw->stack.v_step * depth;
		break;
	case WestGravity:
		x = sw->stack.h_step * depth;
		y = 0;
		break;
	case CenterGravity:
	case SouthGravity:
	case EastGravity:
	default:
		x = y = 0;
		break;
	}
	if (x_return)
		*x_return = sw->stack.highlight_width + x;
	if (y_return)
		*y_return = sw->stack.highlight_height + y;
}

/*
 * Call the split callback, returning True if it is okay
 * to allow the user split the stack. Defaults to True.
 */
static Boolean
StackCallSplitCallback(sw, depth)
	StackWidget sw;
	Cardinal depth;
{
	StackSplitCallbackProc split_proc;

	split_proc = sw->stack.split_callback;
	if (split_proc)
		return (*split_proc)((Widget)sw, NULL, depth);
	else
		return True;
}

/*
 * Call the drop callback, returning True if it is okay
 * to allow the user to drop the cards on the stack.
 * Defaults to True.
 */
static Boolean
StackCallDropCallback(sw, cards, num_cards, source)
	StackWidget sw, source;
	CardID *cards;
	Cardinal num_cards;
{
	StackDropCallbackProc drop_proc;

	drop_proc = sw->stack.drop_callback;
	if (drop_proc)
		return (*drop_proc)((Widget)sw, NULL, cards, num_cards,
			(Widget)source);
	else
		return True;
}

/*
 * Return the drawing extent needed for a particular hstep/vstep
 */
static void
StackExtent(sw, hstep, vstep, width_return, height_return)
	StackWidget sw;
	Dimension hstep, vstep;
	Dimension *width_return, *height_return;
{
	int w, h, index;
	CardObject co;
	Cardinal depth = StackCount(sw);

	HERE();
	if (depth == 0) {
		StackGetPixmapSize(sw, _CARD_PLACE);
		w = sw->stack.card_width[_CARD_PLACE];
		h = sw->stack.card_height[_CARD_PLACE];
	} else {
		co = StackGetCard(sw, depth - 1);
		index = CARDID_TO_INDEX(co->card.card_id);
		if (!sw->stack.card_pixmap[index]) {
			char msg[80];
			snprintf(msg, sizeof msg,
			    "No pixmap for card 0x%0x (%s of %s)",
			    index, CardFaceName(index), CardSuitName(index));
			XtWarning(msg);
			index = _CARD_PLACE;		/* XXX */
		}
		StackGetPixmapSize(sw, index);
		w = sw->stack.card_width[index];
		h = sw->stack.card_height[index];
		switch (sw->stack.gravity) {
		case SouthGravity:
		case NorthGravity:
			h += (depth - 1) * vstep;
			break;
		case EastGravity:
		case WestGravity:
			w += (depth - 1) * hstep;
			break;
		}
	}
	if (width_return) *width_return = w + 2 * sw->stack.highlight_width;
	if (height_return) *height_return = h + 2 * sw->stack.highlight_height;
}

/*
 * Fill the highlight area
 */
void
FillHighlightArea(StackWidget sw)
{
	GC gc = sw->stack.highlight_gc;

	HERE();
	if (sw->stack.highlight_height) {
		XFillRectangle(XtDisplay(sw), XtWindow(sw), gc,
			0, 0, 
			XtWidth(sw), sw->stack.highlight_height);
		XFillRectangle(XtDisplay(sw), XtWindow(sw), gc,
			0, XtHeight(sw) - sw->stack.highlight_height,
			XtWidth(sw), sw->stack.highlight_height);
	}
	if (sw->stack.highlight_width) {
		XFillRectangle(XtDisplay(sw), XtWindow(sw), gc,
			0, sw->stack.highlight_height,
			sw->stack.highlight_width, 
			XtHeight(sw) - 2 * sw->stack.highlight_height);
		XFillRectangle(XtDisplay(sw), XtWindow(sw), gc,
			XtWidth(sw) - sw->stack.highlight_width,
			sw->stack.highlight_height,
			sw->stack.highlight_width, 
			XtHeight(sw) - 2 * sw->stack.highlight_height);
	}
}

/*
 * Clear the highlight area.
 */
void
ClearHighlightArea(StackWidget sw)
{
	HERE();
	if (sw->stack.highlight_width) {
		XClearArea(XtDisplay(sw), XtWindow(sw),
			0, 0, 
			XtWidth(sw), sw->stack.highlight_height, True);
		XClearArea(XtDisplay(sw), XtWindow(sw),
			0, XtHeight(sw) - sw->stack.highlight_height,
			XtWidth(sw), sw->stack.highlight_height, True);
	}
	if (sw->stack.highlight_height) {
		XClearArea(XtDisplay(sw), XtWindow(sw),
			0, sw->stack.highlight_height,
			sw->stack.highlight_width, 
			XtHeight(sw) - 2 * sw->stack.highlight_height, True);
		XClearArea(XtDisplay(sw), XtWindow(sw),
			XtWidth(sw) - sw->stack.highlight_width,
			sw->stack.highlight_height,
			sw->stack.highlight_width, 
			XtHeight(sw) - 2 * sw->stack.highlight_height, True);
	}
}

/*
 * Return local coords to the center of the bottom card in the stack.
 * (used to answer the 'what is this stack over' question)
 */
static void
StackGetBottomCentre(sw, x_return, y_return)
	StackWidget sw;
	Position *x_return, *y_return;
{
	Position x, y, xh, yh;
	CardObject co;
	Cardinal depth = StackCount(sw);
	int index;

	if (!depth)
		x = y = 0;
	else {
		co = StackGetCard(sw, 0);
		index = CARDID_TO_INDEX(co->card.card_id);
		StackGetPixmapSize(sw, index);
		if (!sw->stack.card_pixmap[index])
			index = _CARD_PLACE;		/* hmm. */
		xh = sw->stack.card_width[index] / 2;
		yh = sw->stack.card_height[index] / 2;

		switch (sw->stack.gravity) {
		default:
		case CenterGravity:
			x = sw->stack.highlight_width + xh;
			y = sw->stack.highlight_height + yh;
			break;
		case NorthGravity:
			x = XtHeight(sw) - 
				sw->stack.highlight_height - 
				(depth - 1) * sw->stack.v_step -
				yh;
			y = sw->stack.highlight_height +
				yh;
			break;
		case EastGravity:
			x = sw->stack.highlight_width +
				xh;
			y = XtWidth(sw) - 
				sw->stack.highlight_width - 
				(depth - 1) * sw->stack.h_step -
				yh;
			break;
		case SouthGravity:
			x = sw->stack.highlight_width + 
				xh;
			y = sw->stack.highlight_height +
				(depth - 1) * sw->stack.v_step +
				yh;
			break;
		case WestGravity:
			x = sw->stack.highlight_width +
				(depth - 1) * sw->stack.h_step +
				xh;
			y = sw->stack.highlight_height +
				yh;
			break;
		}
	}
	if (x_return) *x_return = x;
	if (y_return) *y_return = y;
}

/*------------------------------------------------------------
 * Composite methods
 */

/*
 * A child is about to be inserted.
 * We reject all non-card children.
 */
static void
StackInsertChild(Widget w)
{
	StackWidget sw = (StackWidget)w->core.parent;
	XtWidgetProc insert_child;

	HERE();
	XtCheckSubclass(w, cardObjectClass, "Stack: inserted non-Card object");
	if ((insert_child = 
	    StackSuperclass(sw)->composite_class.insert_child) != NULL)
		(*insert_child)(w);
}

/*
 * Called when the list of managed widgets changes
 */
static void
StackChangeManaged(Widget w)
{
	StackWidget sw = (StackWidget)w;
	Dimension width, height;
	Dimension v_step, h_step;

	HERE();

	/* Ask if we can grow as big as possible */
	StackExtent(sw, sw->stack.h_step_max, sw->stack.v_step_max, 
	    &width, &height);
	if (!sw->stack.compact) {
		/* If the widget is already big, then we leave it as is */
		if (width < XtWidth(w)) width = XtWidth(w);
		if (height < XtHeight(w)) height = XtHeight(w);
	}
	if (width && height && (width != XtWidth(w) || height != XtHeight(w))) {
		XtGeometryResult result;
		Dimension almost_width, almost_height;

		result = XtMakeResizeRequest(w, width, height, 
			&almost_width, &almost_height);
	}

	/* Resize anyway - we may need to change our stretch */
	v_step = sw->stack.v_step;
	h_step = sw->stack.h_step;
	StackResize(w);
	if (v_step != sw->stack.v_step ||
	    h_step != sw->stack.h_step) 
		XClearArea(XtDisplay(w), XtWindow(w), 
		    0, 0, w->core.width, w->core.height, True);
}

/*------------------------------------------------------------
 * Core methods
 */

static void
StackClassInitialize()
{
	HERE();

	XtAddConverter(XtRString, XtRGravity, XmuCvtStringToGravity, NULL, 0);
	XtSetTypeConverter(XtRGravity, XtRString, XmuCvtGravityToString,
	    NULL, 0, XtCacheNone, NULL);
}

/*
 * Initialise an instance of a stack widget
 */
static void
StackInitialize(Widget request, Widget w, ArgList args, Cardinal *num_args)
{
	StackWidget sw = (StackWidget)w;
	int index;

	HERE();
	sw->stack.v_step = sw->stack.v_step_max;
	sw->stack.h_step = sw->stack.h_step_max;
	sw->stack.gc = None;
	sw->stack.highlight_gc = None;
	sw->stack.highlighted = False;

	/* Set the pixmap depth to -1 meaning we need to look it up */
	for (index = 0; index < XtNumber(sw->stack.card_width); index++)
		sw->stack.card_depth[index] = -1;

	/* Never have a pixmap for CARD_NONE */
	sw->stack.card_pixmap[CARD_NONE] = NULL;

	StackGetGC(sw);
	if (XtWidth(sw) == 0 || XtHeight(sw) == 0) {
		StackGetPixmapSize(sw, _CARD_PLACE);
		if (XtWidth(sw) == 0)
			XtWidth(sw) = sw->stack.card_width[_CARD_PLACE]
				+ 2 * sw->stack.highlight_width;
		if (XtHeight(sw) == 0)
			XtHeight(sw) = sw->stack.card_height[_CARD_PLACE]
				+ 2 * sw->stack.highlight_height;
	}
}

/*
 * Destroy resources allocated during the stack widget's existence.
 */
static void
StackDestroy(Widget w)
{
	StackWidget sw = (StackWidget)w;

	HERE();

	if (sw->stack.gc) {
		XtReleaseGC(w, sw->stack.gc);
		sw->stack.gc = None;
	}
	if (sw->stack.highlight_gc) {
		XtReleaseGC(w, sw->stack.highlight_gc);
		sw->stack.highlight_gc = None;
	}
}

/*
 * Called when the widget is given an actual window
 */
static void
StackRealize(Widget w, XtValueMask *value_mask, 
	XSetWindowAttributes *attributes)
{
	XtRealizeProc realize;

	HERE();
	if ((realize = StackSuperclass(w)->core_class.realize) != NULL)
		(*realize)(w, value_mask, attributes);
}

/*
 * Draw the contained cards.
 */
static void
StackExpose(Widget w, XEvent *event, Region region)
{
	StackWidget sw = (StackWidget)w;
	int sx, sy, pw, ph, index;
	int v_step = sw->stack.v_step;
	int h_step = sw->stack.h_step;
	Dimension iw = sw->stack.highlight_width;
	Dimension ih = sw->stack.highlight_height;
	CardObject co;
	int i;
	Boolean top;
	XtExposeProc expose;
	Cardinal depth;

	HERE();
	if ((expose = StackSuperclass(w)->core_class.expose) != NULL)
		(*expose)(w, event, region);

	depth = StackCount(sw);
	if (depth == 0) {
		StackGetPixmapSize(sw, _CARD_PLACE);
		StackDrawPixmap(sw, _CARD_PLACE,
		    0, 0,
		    sw->stack.card_width[_CARD_PLACE],
		    sw->stack.card_height[_CARD_PLACE],
		    sw->stack.highlight_width,
		    sw->stack.highlight_height, 
		    region);
	}

	for (i = 0; i < depth; i++) {
		top = (i == depth - 1);

		co = StackGetCard(sw, i);
		index = CARDID_TO_INDEX(co->card.card_id);
		if (!sw->stack.card_pixmap[index]) {
			index = _CARD_PLACE;
			/* XXX this is bad - should warn */
		}
		StackGetPixmapSize(sw, index);
		pw = sw->stack.card_width[index];
		ph = sw->stack.card_height[index];

		switch (sw->stack.gravity) {
		case SouthGravity:
			sx = 0;
			sy = top ? 0 : ph - v_step;
			co->rectangle.x = iw;
			co->rectangle.y = sw->core.height - iw - top 
					? i * v_step + ph
					: (i+1) * v_step;
			co->rectangle.width = pw;
			co->rectangle.height = top ? ph : v_step;
			break;
		case NorthGravity:
			sx = 0;
			sy = 0;
			co->rectangle.x = iw;
			co->rectangle.y = ih + i * v_step;
			co->rectangle.width = pw;
			co->rectangle.height = top ? ph : v_step;
			break;
		case WestGravity:
			sx = 0;
			sy = 0;
			co->rectangle.x = iw + i * h_step;
			co->rectangle.y = ih ;
			co->rectangle.width = top ? pw : h_step;
			co->rectangle.height = ph;
			break;
		case EastGravity:
			sx = top ? 0 : pw - h_step;
			sy = 0;
			co->rectangle.x = sw->core.width - iw - top 
					? i * h_step + pw
					: (i+1) * h_step;
			co->rectangle.y = ih;
			co->rectangle.width = top ? pw : h_step;
			co->rectangle.height = ph;
			break;
		default:
		case CenterGravity:
			sx = 0;
			sy = 0;
			co->rectangle.x = iw;
			co->rectangle.y = ih;
			co->rectangle.width = top ? pw : 0;
			co->rectangle.height = top ? ph : 0;
			break;
		}
		StackDrawPixmap(sw, index, sx, sy, 
		    co->rectangle.width, co->rectangle.height,
		    co->rectangle.x, co->rectangle.y,
		    region);
	}
	if (sw->stack.highlighted)
		FillHighlightArea(sw);
}

/*
 * Called when a resource value is changed
 */
static Boolean
StackSetValues(Widget current, Widget request, Widget w, 
	ArgList args, Cardinal *num_args)
{
	StackWidget rcurrent = (StackWidget)current;
	/*StackWidget rrequest = (StackWidget)request;*/
	StackWidget rw = (StackWidget)w;
	int i, index;
	Boolean redisplay = False;

	HERE();

	/*
	 * If the vstep/hstep limits change, we should invoke resize
	 * to change the steps.
	 */
	if (rcurrent->stack.h_step_max != rw->stack.h_step_max ||
	    rcurrent->stack.v_step_max != rw->stack.v_step_max ||
	    rcurrent->stack.h_step_min != rw->stack.h_step_min ||
	    rcurrent->stack.v_step_min != rw->stack.v_step_min ||
	    rcurrent->stack.highlight_width != rw->stack.highlight_width ||
	    rcurrent->stack.highlight_height != rw->stack.highlight_height)
	{
		redisplay = True;
	}

	/*
	 * If the gravity changes, we need to redraw
	 */
	if (rcurrent->stack.gravity != rw->stack.gravity)
	{
		if (rw->stack.gravity != CenterGravity &&
		    rw->stack.gravity != NorthGravity &&
		    rw->stack.gravity != SouthGravity &&
		    rw->stack.gravity != EastGravity &&
		    rw->stack.gravity != WestGravity)
		{
			rw->stack.gravity = CenterGravity;
			XtWarning("Stack: setting gavity to Center");
		}
		redisplay = True;
	}

	/*
	 * If a card's pixmap changed, we should invalidate its cached
	 * width/height/depth
	 */
	for (index = 0; index < XtNumber(rw->stack.card_pixmap); index++)
		if (rcurrent->stack.card_pixmap[index] !=
		    rw->stack.card_pixmap[index])
		{
			rw->stack.card_depth[index] = -1;
			redisplay = True;
		}

	/*
	 * If any of the visible pixmaps changed, we should
	 * redraw.
	 */
	i = 0;
	if (rw->stack.gravity == CenterGravity &&
	    rw->composite.num_children > 0)
		i = rw->composite.num_children - 1;
	for (; i < rw->composite.num_children; i++)
	{
		CardObject co = (CardObject)rw->composite.children[i];
		if (!XtIsManaged((Widget)co))
			continue;
		index = CARDID_TO_INDEX(co->card.card_id);
		if (rcurrent->stack.card_pixmap[index] != 
		    rw->stack.card_pixmap[index])
		{
			redisplay = True;
		}
	}

	if (redisplay)
		StackResize(w);
	return redisplay;
}

/*
 * Notification that this widget's size has changed (and stabilised).
 * We compute the best hstep/vstep here.
 */
static void
StackResize(Widget w)
{
	StackWidget sw = (StackWidget)w;
	XtWidgetProc resize;
	int hstep, vstep;
	Dimension width, height;

	HERE();

	/* Call the superclass's resize */
	if ((resize = StackSuperclass(w)->core_class.resize))
		(*resize)(w);

	/* Start with the largest step size */
	hstep = sw->stack.h_step_max;
	vstep = sw->stack.v_step_max;

	/* Decrease the step size until we fit or hit the limit */
	if (sw->stack.gravity == SouthGravity ||
	    sw->stack.gravity == NorthGravity)
        {
	    StackExtent(sw, hstep, vstep, &width, &height);
	    while (vstep > sw->stack.v_step_min && height > XtHeight(w)) {
		vstep--;
	        StackExtent(sw, hstep, vstep, &width, &height);
	    }
	}
	else if (sw->stack.gravity == EastGravity ||
	         sw->stack.gravity == WestGravity)
        {
	    StackExtent(sw, hstep, vstep, &width, &height);
	    while (hstep > sw->stack.h_step_min && width > XtWidth(w)) {
		hstep--;
	        StackExtent(sw, hstep, vstep, &width, &height);
	    }
	}
	sw->stack.h_step = hstep;
	sw->stack.v_step = vstep;
}

/*
 * Someone is requesting our preferred size, given a constraining box.
 */
static XtGeometryResult 
StackQueryGeometry(w, request, preferred_return)
	Widget w;
	XtWidgetGeometry *request;
	XtWidgetGeometry *preferred_return;
{
	StackWidget sw = (StackWidget)w;
	Dimension max_width, max_height;
	Dimension min_width, min_height;

	HERE();

	/* Any stacking order or location is fine */
	if ((request->request_mode & (CWWidth|CWHeight)) == 0)
		return XtGeometryYes;

	/* Figure out our biggest size */
	StackExtent(sw, sw->stack.h_step_max, sw->stack.v_step_max, 
	    &max_width, &max_height);
	preferred_return->width = max_width;
	preferred_return->height = max_height;
	preferred_return->request_mode = CWWidth|CWHeight;

	/* If we already fit inside the request geometry, then good. */
	if ((((request->request_mode & CWWidth) == 0) ||
	      request->width >= max_width) &&
	    (((request->request_mode & CWHeight) == 0) ||
	      request->height >= max_height))
		return XtGeometryYes;

	/* 
	 * If our smallest extent fits inside the request, then anything
	 * in between will be ok (if tight)
	 */
	StackExtent(sw, sw->stack.h_step_min, sw->stack.v_step_min, 
	    &min_width, &min_height);
	if ((((request->request_mode & CWWidth) == 0) ||
	      request->width >= min_width) &&
	    (((request->request_mode & CWHeight) == 0) ||
	      request->height >= min_height))
		return XtGeometryAlmost;

	/* Otherwise its just too tight for us */
	return XtGeometryNo;
}

/*------------------------------------------------------------
 * Action Implementation
 */

static void
Highlight(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	StackWidget sw = (StackWidget)w;

	HERE();
	if (*num_params != (Cardinal)0)
		XtWarning("Too many parameters to hightlight()");

	if (!sw->stack.highlighted) {
		sw->stack.highlighted = True;
		if (XtIsRealized(w))
			FillHighlightArea(sw);
	}
}

static void
Unhighlight(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	StackWidget sw = (StackWidget)w;

	HERE();
	if (*num_params != (Cardinal)0)
		XtWarning("Too many parameters to hightlight()");

	if (sw->stack.highlighted) {
		sw->stack.highlighted = False;
		if (XtIsRealized(w))
			ClearHighlightArea(sw);
	}
}

/*------------------------------------------------------------
 * Drag & drop implementation
 */

static Widget popup;
static Widget popup_stack;
static Widget popup_toplevel = NULL;
static Widget popup_source;
static Position popup_xoff, popup_yoff;
static XtEventDispatchProc popup_old_dispatch_m, popup_old_dispatch_r;
static Widget popup_over = NULL;
static Widget popup_target = NULL;

/*
 * Create the popup used for dragging and dropping
 */
Widget
StackCreateDragAndDropPopup(Widget toplevel, String popup_name)
{
	Arg args[1];
	Cardinal argc;

	if (popup_toplevel == toplevel)
		return popup;
	if (popup_toplevel != NULL)
		XtError("StackCreateDragAndDropPopup: too many toplevels");
	popup_toplevel = toplevel;

	/*
	 * Create the popup window, and allow the cards inside
	 * to dictate its size.
	 */
	argc = 0;
	XtSetArg(args[argc], XtNallowShellResize, True); argc++;

	popup = XtCreatePopupShell(popup_name, overrideShellWidgetClass, 
		toplevel, args, argc);

	/*
	 * Create the stack within the popup window.
	 * We set its compact field to true so that it takes minimal space
	 */
	argc = 0;
	XtSetArg(args[argc], XtNcompact, True); argc++;

	popup_stack = XtCreateManagedWidget("stack", stackWidgetClass,
		popup, args, argc);
	return popup;

	/*
	 * XXX - need to add action/translation to popup so that Escape
	 * aborts the drag?
	 */
}

static void
StartDrag(w, event, params, param_count)
	Widget w;
	XEvent *event;
	String *params;
	Cardinal *param_count;
{
	Position x, y;
	Position cardx, cardy;
	StackWidget sw = (StackWidget)w;
	Cardinal ncards;
	int startcard;
	Cardinal stack_depth;

	/* No toplevel? */
	if (!popup_toplevel)
		return;

	/* Is something being dragged already? */
	if (popup_source)
		return;

	XtCheckSubclass(w, stackWidgetClass, "Drag: dragging non-stack");

	/* How many cards are we trying to drag? */
	ncards = StackCountSelected(w, event->xbutton.x, event->xbutton.y);
	if (ncards == 0)
		return;

	/* Are we allowed to split this stack this way? */
	if (!StackCallSplitCallback(sw, ncards))
		return;

	/*
	 * Find the global position of the top left corner of the 
	 * lowest selected card
	 */
	stack_depth = StackCount(sw);
	startcard = stack_depth - ncards;
	StackPartOrigin(sw, startcard, &cardx, &cardy);
	XtTranslateCoords(w, cardx, cardy, &x, &y);

	/*
	 * Figure out the pixel offset into the selected card
	 * that the cursor will be 'attached' to
	 */
	popup_xoff = event->xbutton.x - cardx;
	popup_yoff = event->xbutton.y - cardy;
	popup_source = w;
	popup_target = NULL;

	/* Move cards from the card source into the popup */
	StackMove(popup_stack, w, ncards);

	/* Move the popup to where it should begin */
	XtMoveWidget(popup, x, y);

	/* And show it */
	XtPopup(popup, XtGrabExclusive);

	/* Start grabbing pointer events relative to toplevel */
	XtGrabPointer(popup_toplevel,
		False,			/* owner_events */
		PointerMotionMask 	/* event_mask */
		|ButtonReleaseMask,
		GrabModeAsync,		/* pointer_mode */
		GrabModeAsync,		/* keyboard_mode */
		None,			/* confine_to */
		None,			/* cursor */
		event->xbutton.time	/* time */
	);

	popup_old_dispatch_m = XtSetEventDispatcher(XtDisplay(popup_toplevel), 
		MotionNotify, PopupEventHandler);
	popup_old_dispatch_r = XtSetEventDispatcher(XtDisplay(popup_toplevel), 
		ButtonRelease, PopupEventHandler);
}

static void
AbortDrag(w, event, params, param_count)
	Widget w;
	XEvent *event;
	String *params;
	Cardinal *param_count;
{
	if (popup_source)
		PopupStopDrag(event, True);
}

static void
Hit(w, event, params, param_count)
	Widget w;
	XEvent *event;
	String *params;
	Cardinal *param_count;
{
	StackWidget sw = (StackWidget)w;

	XtCallCallbackList(w, sw->stack.hit_notify, (XtPointer) NULL);
}

/*
 * A motion/release event has occurred relative to toplevel.
 * We track and move the popup window, and see if we need to drop.
 */
static Boolean
PopupEventHandler(event)
	XEvent *event;
{
	Position x, y;
	Widget w;

	switch (event->type) {
	case MotionNotify:
		XtTranslateCoords(popup_toplevel, event->xmotion.x, 
			event->xmotion.y, &x, &y);
		XtMoveWidget(popup, x - popup_xoff, y - popup_yoff);
		XtTranslateCoords(popup_toplevel, event->xmotion.x, 
			event->xmotion.y, &x, &y);
		/*
		 * Find the center of the lowest card on the
		 * carried stack in global coordinates - and use
		 * that as the search point (rather than the cursor)
		 */
		StackGetBottomCentre((StackWidget)popup_stack, &x, &y);
		x = event->xmotion.x - popup_xoff + x;
		y = event->xmotion.y - popup_yoff + y;
		w = FindWidget(popup_toplevel, x, y);
		if (w != popup_over) {
			if (popup_target) {
				PopupLeave(popup_target, event);
				popup_target = NULL;
			}
			popup_over = w;
			if (popup_over) {
				if (PopupCanEnter(popup_over, event)) {
					popup_target = popup_over;
					PopupEnter(popup_target, event);
				}
			} 
		}
		return False;
	case ButtonRelease:
		if (event->xbutton.button == 1) {
			PopupStopDrag(event, False);
			return True;
		}
		/* fallthrough */
	default:
		return False;
	}
}

/*
 * Completes the popup drag action.
 */
static void
PopupStopDrag(event, cancelled)
	XEvent *event;
	Boolean cancelled;
{
	if (!popup_target)
		cancelled = True;
	if (popup_target) 
		PopupLeave(popup_over, event);
	if (cancelled) 
		StackMove(popup_source, popup_stack, 
			StackDepth(popup_stack));
	if (popup_target && !cancelled) {
		StackDropEvent ev;
		Cardinal count;
		StackWidget sw = (StackWidget)popup_target;

		count = StackDepth(popup_stack);
		ev.source = popup_source;
		ev.target = popup_target;
		ev.count = count;

		StackMove(popup_target, popup_stack, count);
		XtCallCallbackList(popup_target, sw->stack.drop_notify, 
			(XtPointer) &ev);
	}
	popup_target = NULL;

	if (popup_old_dispatch_m) {
		XtSetEventDispatcher(XtDisplay(popup), 
		    MotionNotify, popup_old_dispatch_m);
		XtSetEventDispatcher(XtDisplay(popup), 
		    ButtonRelease, popup_old_dispatch_r);
		popup_old_dispatch_m = NULL;
		popup_old_dispatch_r = NULL;
	}
	XtUngrabPointer(popup_toplevel, event->xbutton.time);
	popup_source = NULL;
	XtPopdown(popup);
}

/*
 * Find the widget visible at location x,y relative to origin of w
 * Returns NULL if no widget.
 * If w is not composite and x,y is inside w, then w is returned.
 * (Non-widget children of composites are not searched)
 */
static Widget
FindWidget(Widget w, int x, int y)
{
	Widget r;
	int i;
	CompositeWidget cw;

	if (x < 0 || x >= w->core.width ||
	    y < 0 || y >= w->core.height)
		return NULL;
	if (XtIsComposite(w)) {
		cw = (CompositeWidget)w;
		for (i = 0; i < cw->composite.num_children; i++) {
			Widget subw = cw->composite.children[i];
			if (XtIsWidget(subw) && 
			    XtIsManaged(subw) &&
			    (r = FindWidget(subw, 
				x - subw->core.x - subw->core.border_width,
				y - subw->core.y - subw->core.border_width))
			   )
				return r;
		}
	}
	return w;
}

/*
 * Return true if the popup moves over an enterable widget.
 * (Normally only stack widgets that would allow a drop are enterable)
 * (The source stack is not considered as enterable)
 */
static Boolean
PopupCanEnter(Widget w, XEvent *e)
{
	if (!XtIsSubclass(w, stackWidgetClass))
		return False;
	if (w == popup_source)
		return False;
	return PopupCanDropCards(w);
}

/*
 * Called when the dragged popup enters an enterable widget
 */
static void
PopupEnter(Widget w, XEvent *e)
{
	if (XtIsSubclass(w, stackWidgetClass))
		XtCallActionProc(w, "highlight", e, NULL, 0);
}

/*
 * Called when the dragged popup leaves an enterred widget
 */
static void
PopupLeave(Widget w, XEvent *e)
{
	if (XtIsSubclass(w, stackWidgetClass))
		XtCallActionProc(w, "unhighlight", e, NULL, 0);
}

/*
 * Try dropping the cards from the popup into the widget.
 * Returns true if ok.
 */
static Boolean
PopupCanDropCards(Widget w)
{
	Boolean result;
	CardID *cards;
	Cardinal num_cards;
	int i;

	HERE();
	if (!XtIsSubclass(w, stackWidgetClass))
		return False;
	num_cards = StackDepth(popup_stack);
	cards = (CardID *)ALLOCATE_LOCAL(num_cards * sizeof (CardID));
	for (i = 0; i < num_cards; i++)
		cards[i] = StackPeek(popup_stack, i);
	result = StackCallDropCallback((StackWidget)w,
	    cards, num_cards, (StackWidget)popup_stack);	/* XXX */
	DEALLOCATE_LOCAL(cards);
	return result;
}
